import structlog
from django.conf import settings

logger = structlog.get_logger(__name__)


def patch_celery_default_queue(celery_app):
    try:
        from celery import bootsteps
        from kombu import Exchange, Queue
        from kombu.common import QoS
    except ImportError:
        logger.exception("Celery is not installed")
        return

    if "amqp://" in settings.CELERY_BROKER_URL:
        # The following is only needed when using RabbitMQ
        # See https://github.com/celery/celery/issues/6067#issuecomment-724003961
        # only needed until a new major version of Celery is released

        class NoChannelGlobalQoS(bootsteps.StartStopStep):
            requires = {"celery.worker.consumer.tasks:Tasks"}

            def start(self, c):
                qos_global = False
                c.connection.default_channel.basic_qos(
                    0,
                    c.initial_prefetch_count,
                    qos_global,
                )

                def set_prefetch_count(prefetch_count):
                    return c.task_consumer.qos(
                        prefetch_count=prefetch_count,
                        apply_global=qos_global,
                    )

                c.qos = QoS(set_prefetch_count, c.initial_prefetch_count)

        # Only use the custom QoS when using RabbitMQ
        # Make sure to create the queue with the x-queue-type quorum argument as this ensure that the queue is durable
        # and replicated across multiple nodes.
        celery_app.steps["consumer"].add(NoChannelGlobalQoS)
        celery_app.conf.task_queues = [
            Queue(
                settings.CELERY_TASK_DEFAULT_QUEUE,
                Exchange(settings.CELERY_TASK_DEFAULT_QUEUE),
                routing_key=settings.CELERY_TASK_DEFAULT_QUEUE,
                queue_arguments={'x-queue-type': 'quorum'},
            ),
        ]



"""
Settings for Payla Utils are all namespaced in the PAYLA_UTILS setting.
For example your project's `settings.py` file might look like this:
PAYLA_UTILS = {
    'APP_NAME': 'My App',
    # Used for json logging
    'MICROSERVICE_NAME: 'myapp',
    # stage, playground, prod ...
    'ENVIRONMENT': ENVIRONMENT,
    'INITIAL_FIXTURES': [
        os.path.join(BASE_DIR, 'testapp', 'fixtures', 'users.json'),
    ],
    'SERVER_IP': '192.168.1.4',
    'REQUEST_ID_HEADER': 'X-Request-ID',
    'RUN_EXTRA_COMMANDS': ['loadinitialusers', 'setup_something'],
    'LOCAL_DJANGO_ADMIN_PASSWORD': os.environ.get('LOCAL_DJANGO_ADMIN_PASSWORD', 'admin'),
    'USE_PGTRIGGER': False,
    'USE_HISTORICAL_MODELS': False,
    'HISTORICAL_IGNORE_FIELD_NAME': [],
    'ENV_THEMES': {
        'local.dev': {
            'title_color': '#000000',
            'css_header_background_color': '#ffffff',
            'env_color': '#00cb38',
            'css_header_text_color': '#000000',
            'css_header_link_color': '#000000',
            'css_header_link_hover_color': '#1e00ac',
            'css_module_background_color': '#ababab',
            'css_module_background_selected_color': '#e3e3e3',
            'css_module_text_color': '#000000',
            'css_module_link_color': '#000000',
            'css_module_link_hover_color': '#3255fe',
            'css_generic_link_color': '#000000',
            'css_save_button_background_color': '#6d6d6d',
            'css_save_button_background_hover_color': '#4a4a4a',
        },
        'stage': {
            'title_color': '#ffffff',
            'env_color': '#ffcb38',
            'css_header_background_color': '#ff9722',
            'css_header_text_color': '#ffffff',
            'css_header_link_color': '#ffffff',
            'css_header_link_hover_color': '#41aad1',
            'css_module_background_color': '#ca6a00',
            'css_module_background_selected_color': '#ffffff',
            'css_module_text_color': '#ffffff',
            'css_module_link_color': '#ffffff',
            'css_module_link_hover_color': '#41aad1',
            'css_generic_link_color': '#000000',
            'css_save_button_background_color': '#ca6a00',
            'css_save_button_background_hover_color': '#ff9722',
        },
        'playground': {
            'title_color': '#ffffff',
            'env_color': '#00cb38',
            'css_header_background_color': '#09137a',
            'css_header_text_color': '#ffffff',
            'css_header_link_color': '#ffffff',
            'css_header_link_hover_color': '#1e00ac',
            'css_module_background_color': '#0020bf',
            'css_module_background_selected_color': '#e3e3e3',
            'css_module_text_color': '#ffffff',
            'css_module_link_color': '#ffffff',
            'css_module_link_hover_color': '#69c2cc',
            'css_generic_link_color': '#000000',
            'css_save_button_background_color': '#0038ff',
            'css_save_button_background_hover_color': '#02208b',
        },
        'prod': {
            'title_color': '#ffffff',
            'env_color': '#00cb38',
            'css_header_background_color': '#720606',
            'css_header_text_color': '#ffffff',
            'css_header_link_color': '#ffffff',
            'css_header_link_hover_color': '#1e00ac',
            'css_module_background_color': '#e73f41',
            'css_module_background_selected_color': '#e3e3e3',
            'css_module_text_color': '#ffffff',
            'css_module_link_color': '#ffffff',
            'css_module_link_hover_color': '#5f0000',
            'css_generic_link_color': '#000000',
            'css_save_button_background_color': '#720606',
            'css_save_button_background_hover_color': '#4a4a4a',
            # APP_NAME will be replaced by the correct app name set in payla utils settings
            'title': 'APP_NAME',
        },
    }
}
This module provides the `payla_utils_setting` object, that is used to access
Payla Utils settings, checking for user settings first, then falling
back to the defaults.
"""

from __future__ import annotations

from django.conf import settings

# Import from `django.core.signals` instead of the official location
# `django.test.signals` to avoid importing the test module unnecessarily.
from django.core.signals import setting_changed

DEFAULTS: dict = {
    'APP_NAME': None,
    'MICROSERVICE_NAME': None,
    'ENVIRONMENT': None,
    'INITIAL_FIXTURES': [],
    'RUN_EXTRA_COMMANDS': [],
    'LOCAL_DJANGO_ADMIN_PASSWORD': None,
    'REQUEST_ID_HEADER': None,
    'SERVER_IP': None,
    'USE_PGTRIGGER': False,
    'USE_HISTORICAL_MODELS': False,
    'HISTORICAL_IGNORE_MODELS': [],
    'HISTORICAL_IGNORE_FIELD_NAME': [],
    'GROUPS_PERMISSIONS_FILE_PATH': None,
    'GROUPS_USERS_FILE_PATH': None,
    'USER_PASSWORDS': {},
    'ENV_THEMES': {
        'local.dev': {
            'title_color': '#000000',
            'css_header_background_color': '#ffffff',
            'env_color': '#00cb38',
            'css_header_text_color': '#000000',
            'css_header_link_color': '#000000',
            'css_header_link_hover_color': '#1e00ac',
            'css_module_background_color': '#ababab',
            'css_module_background_selected_color': '#e3e3e3',
            'css_module_text_color': '#000000',
            'css_module_link_color': '#000000',
            'css_module_link_hover_color': '#3255fe',
            'css_generic_link_color': '#000000',
            'css_save_button_background_color': '#6d6d6d',
            'css_save_button_background_hover_color': '#4a4a4a',
        },
        'stage': {
            'title_color': '#ffffff',
            'env_color': '#ffcb38',
            'css_header_background_color': '#ff9722',
            'css_header_text_color': '#ffffff',
            'css_header_link_color': '#ffffff',
            'css_header_link_hover_color': '#41aad1',
            'css_module_background_color': '#ca6a00',
            'css_module_background_selected_color': '#ffffff',
            'css_module_text_color': '#ffffff',
            'css_module_link_color': '#ffffff',
            'css_module_link_hover_color': '#41aad1',
            'css_generic_link_color': '#000000',
            'css_save_button_background_color': '#ca6a00',
            'css_save_button_background_hover_color': '#ff9722',
        },
        'playground': {
            'title_color': '#ffffff',
            'env_color': '#00cb38',
            'css_header_background_color': '#09137a',
            'css_header_text_color': '#ffffff',
            'css_header_link_color': '#ffffff',
            'css_header_link_hover_color': '#1e00ac',
            'css_module_background_color': '#0020bf',
            'css_module_background_selected_color': '#e3e3e3',
            'css_module_text_color': '#ffffff',
            'css_module_link_color': '#ffffff',
            'css_module_link_hover_color': '#69c2cc',
            'css_generic_link_color': '#000000',
            'css_save_button_background_color': '#0038ff',
            'css_save_button_background_hover_color': '#02208b',
        },
        'prod': {
            'title_color': '#ffffff',
            'env_color': '#00cb38',
            'css_header_background_color': '#720606',
            'css_header_text_color': '#ffffff',
            'css_header_link_color': '#ffffff',
            'css_header_link_hover_color': '#1e00ac',
            'css_module_background_color': '#e73f41',
            'css_module_background_selected_color': '#e3e3e3',
            'css_module_text_color': '#ffffff',
            'css_module_link_color': '#ffffff',
            'css_module_link_hover_color': '#5f0000',
            'css_generic_link_color': '#000000',
            'css_save_button_background_color': '#720606',
            'css_save_button_background_hover_color': '#4a4a4a',
            'title': 'APP_NAME',
        },
    },
}


class PaylaUtilsSettings:
    def __init__(self, defaults: dict | None = None) -> None:
        self.defaults = defaults or DEFAULTS
        self._cached_attrs: set = set()

    @property
    def user_settings(self) -> dict:
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'PAYLA_UTILS', {})
        return self._user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError(f"Invalid setting: '{attr}'")

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self) -> None:
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


payla_utils_settings = PaylaUtilsSettings(DEFAULTS)


def reload_payla_utils_settings(*args, **kwargs) -> None:
    setting = kwargs['setting']
    if setting == 'PAYLA_UTILS':
        payla_utils_settings.reload()


setting_changed.connect(reload_payla_utils_settings)



from logging import LogRecord

from sentry_sdk.integrations.logging import LoggingIntegration

_IGNORED_LOGGERS = set()


def ignore_logger(logger_name: str) -> None:
    _IGNORED_LOGGERS.add(logger_name)


class PaylaLoggingIntegration(LoggingIntegration):
    def _handle_record(self, record: LogRecord) -> None:
        # This match upper logger names, e.g. "celery" will match "celery.worker"
        # or "celery.worker.job"
        if record.name in _IGNORED_LOGGERS or record.name.split(".")[0] in _IGNORED_LOGGERS:
            return
        super()._handle_record(record)



import sys

import structlog
from django.utils.log import DEFAULT_LOGGING


def get_default_logging_conf(log_level: str, formatter: str, formatter_std_lib: str, own_apps: list[str]) -> dict:
    formatters = {
        'verbose': {
            'format': '%(asctime)s %(levelname)s %(name)s %(message)s',
        },
        'django.server': DEFAULT_LOGGING['formatters']['django.server'],
        'json_formatter': {
            '()': 'payla_utils.logging.logformatter.LogFormatter',
            'format': '%(timestamp)s %(level)s %(name)s %(message)s %(pathname)s:%(lineno)d',
        },
        "plain_console": {
            "()": structlog.stdlib.ProcessorFormatter,
            "processor": structlog.dev.ConsoleRenderer(event_key='message'),
        },
        "plain_console_std_lib": {
            # std library + 3rd party loggers don't use structlog and therefore don't run through the processors
            "processor": structlog.dev.ConsoleRenderer(event_key='event'),
        },
        "key_value": {
            "()": structlog.stdlib.ProcessorFormatter,
            "processor": structlog.processors.KeyValueRenderer(
                key_order=['microservice', 'timestamp', 'level', 'event', 'logger']
            ),
        },
    }

    if formatter not in formatters or formatter_std_lib not in formatters:
        raise NotImplementedError("formatter not supported")

    return {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': formatters,
        'handlers': {
            'console': {
                'level': log_level,
                'class': 'logging.StreamHandler',
                'formatter': formatter,
                'stream': sys.stdout,
            },
            'console_std_lib': {
                'level': log_level,
                'class': 'logging.StreamHandler',
                'formatter': formatter_std_lib,
                'stream': sys.stdout,
            },
            'django.server': DEFAULT_LOGGING['handlers']['django.server'],
        },
        "root": {
            "level": log_level,
            "handlers": ["console_std_lib"],
            "propagate": False,
        },
        'loggers': {
            '': {
                'level': log_level,
                'handlers': ['console_std_lib'],
                "propagate": False,
            },
            'django.request': {
                'level': log_level,
                'handlers': ['console_std_lib'],
                "propagate": False,
            },
            "django.security.DisallowedHost": {
                'level': log_level,
                'handlers': ['console_std_lib'],
                "propagate": False,
            },
            'django.server': DEFAULT_LOGGING['loggers']['django.server'],
            "celery.task": {
                "handlers": ["console_std_lib"],
                "level": log_level,
                "propagate": False,
            },
            "celery": {
                "handlers": ["console_std_lib"],
                "level": log_level,
                "propagate": False,
            },
            # Setup loggers for each app
            **{
                app.split('.', maxsplit=1)[0]: {
                    "handlers": ["console"],
                    "level": log_level,
                    "propagate": False,
                }
                for app in own_apps
            },
        },
    }



from __future__ import annotations

import contextlib
from functools import lru_cache
from typing import Any

import pgtrigger
from django.conf import settings
from django.contrib.admin.sites import all_sites
from django.core.checks import Error
from django.db import models
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from payla_utils.settings import payla_utils_settings


class PaylaQuerySet(models.QuerySet):
    """
    This custom QuerySet with get_or_none method
    """

    def get_or_none(self, *args: Any, **kwargs: Any) -> PaylaModel | None:
        """
        Use this instead of Queryset.get to either get an object if it exists or
        None if it doesn't. This is a shortcut to avoid having to catch DoesNotExist
        exceptions.

        Returns:
            PaylaModel | None: Either the model if it exists or None if it doesn't.
        """
        try:
            return self.get(*args, **kwargs)
        except self.model.DoesNotExist:
            return None

    def update(self, *args, **kwargs) -> int:
        """
        Override Queryset.update to make sure that modified_at is also updated when
        performing multi-row updates.

        Returns:
            int: number of rows updated
        """
        if 'modified_at' not in kwargs:
            kwargs['modified_at'] = now()
        return super().update(*args, **kwargs)


class PaylaModelMeta(models.base.ModelBase):
    """
    Metaclass for PaylaModel
    """

    def __new__(cls, name, bases, attrs):
        new_class = super().__new__(cls, name, bases, attrs)
        if payla_utils_settings.USE_PGTRIGGER and not new_class._meta.abstract:
            pgtrigger.register(
                pgtrigger.Trigger(
                    name='update_modified_at',
                    operation=pgtrigger.Update,
                    # Trigger BEFORE updates
                    when=pgtrigger.Before,
                    # Update modified_at field to current time
                    func="NEW.modified_at = clock_timestamp();RETURN NEW;",
                    # Trigger only if something changed in the row, but not the modified_at field
                    condition=pgtrigger.Condition(
                        'OLD.* IS DISTINCT FROM NEW.* AND OLD.modified_at IS NOT DISTINCT FROM NEW.modified_at'
                    ),
                )
            )(new_class)

        if payla_utils_settings.USE_HISTORICAL_MODELS:
            try:
                from simple_history import register
                from simple_history.exceptions import MultipleRegistrationsError
            except ImportError as exc:
                raise ImportError(
                    'simple_history must be installed in order to use PaylaModel with USE_HISTORICAL_MODELS=True'
                ) from exc

            records_config = {
                'inherit': True,
            }
            is_abstract = getattr(new_class._meta, 'abstract', False)
            is_proxy = getattr(new_class._meta, 'proxy', False)
            ignore = new_class._meta.label in payla_utils_settings.HISTORICAL_IGNORE_MODELS
            if not is_abstract and not is_proxy and not ignore:
                # only register historical models for subclasses of PaylaModel

                # exclude globally ignored fields: e.g api_id ...
                for field_to_exclude in payla_utils_settings.HISTORICAL_IGNORE_FIELD_NAME:
                    if field_to_exclude in new_class.__dict__:
                        records_config.setdefault('excluded_fields', []).append(field_to_exclude)

                with contextlib.suppress(MultipleRegistrationsError):
                    register(new_class, **records_config)

        return new_class


class PaylaModel(models.Model, metaclass=PaylaModelMeta):
    created_at = models.DateTimeField(verbose_name=_('Created At'), auto_now_add=True)
    modified_at = models.DateTimeField(verbose_name=_('Last modified'), auto_now=True)

    objects = PaylaQuerySet.as_manager()

    class Meta:
        abstract = True
        indexes = [
            models.Index(fields=['created_at']),
            models.Index(fields=['modified_at']),
        ]

    def save(self, *args, **kwargs):
        """
        Overriding the save method in order to make sure that
        modified field is updated even if it is not given as
        a parameter to the update field argument.
        """
        update_fields = kwargs.get('update_fields')
        if update_fields:
            kwargs['update_fields'] = set(update_fields).union({'modified_at'})

        super().save(*args, **kwargs)

    @classmethod
    @lru_cache(maxsize=1)
    def get_registered_models(cls) -> set[models.Model]:
        registered_models = set()
        for admin_site in all_sites:
            for model, admin in admin_site._registry.items():
                registered_models.add(model)
                for inline in admin.inlines:
                    registered_models.add(inline.model)
        return registered_models

    @classmethod
    def check(cls, **kwargs):
        errors = super().check(**kwargs)
        if hasattr(cls, 'objects') and not issubclass(cls.objects._queryset_class, PaylaQuerySet):
            errors.append(
                Error(
                    f'Model {cls.__name__} objects must be a subclass of PaylaQuerySet',
                    hint=f'Your custom queryset class for model {cls.__name__} must be extending from PaylaQueryset',
                    obj=cls,
                    id='payla_utils.E001',
                )
            )

        if cls not in cls.get_registered_models():
            errors.append(
                Error(
                    f"Model {cls.__name__} does not have an admin",
                    hint=f"Register the model {cls.__name__} in the admin",
                    obj=cls,
                    id='payla_utils.E002',
                )
            )

        for field in cls._meta.get_fields():
            if field.concrete and (field.null and not field.blank):
                errors.append(
                    Error(
                        f"Field {field.name} of model {cls.__name__} is nullable but not blank",
                        hint=f"Either make field {field.name} non-nullable or blank",
                        obj=cls,
                        id='payla_utils.E003',
                    )
                )

        if payla_utils_settings.USE_PGTRIGGER and 'pgtrigger' not in settings.INSTALLED_APPS:
            errors.append(
                Error(
                    'pgtrigger must be installed in order to use PaylaModel with USE_PGTRIGGER=True',
                    hint='Add pgtrigger to INSTALLED_APPS',
                    obj=cls,
                    id='payla_utils.E004',
                )
            )

        if payla_utils_settings.USE_HISTORICAL_MODELS and 'simple_history' not in settings.INSTALLED_APPS:
            errors.append(
                Error(
                    'simple_history must be installed in order to use PaylaModel with USE_HISTORICAL_MODELS=True',
                    hint='Add simple_history to INSTALLED_APPS',
                    obj=cls,
                    id='payla_utils.E005',
                )
            )

        return errors



from rest_framework.permissions import DjangoModelPermissions


class DjangoActionPermissions(DjangoModelPermissions):
    """Customize permission logic based on request, action and object.

    For every viewset with `queryset` you want to override some custom logic
    for particular action, you should define such attribute for `viewset`.
    Viewset permission at common action level:
    ```
        perms_map_action = {
            '<action>': [
                <str> | <function(user, view, obj=None)>,
                ...
            ],
            ...
        }
    ```

    Viewset without queryset should specify the model_cls
    ```
    model_class = ModelName
    ```

    To enforce primary method action i.e for post request, check app.add_modelname permission and the specified
    perms_map_action
    ```
    include_model_default_method_permission = True
    ```
    """

    perms_map_action = {
        "retrieve": ["%(app_label)s.view_%(model_name)s"],
        "list": ["%(app_label)s.view_%(model_name)s"],
        "create": ["%(app_label)s.add_%(model_name)s"],
        "update": ["%(app_label)s.change_%(model_name)s"],
        "destroy": ["%(app_label)s.delete_%(model_name)s"],
    }

    def generate_perms_map_action(self, view):
        action = view.action
        if action:
            action = action.replace("partial_", "")  # Treat partial updates the same as normal updates
            if self.perms_map_action.get(action):
                return
            perm_str = f"%(app_label)s.{action}_%(app_label)s"
            permission_base = getattr(view, "permission_base", "")
            if permission_base:
                perm_str = f"%(app_label)s.{action}_{permission_base}"

            self.perms_map_action[action] = [perm_str]

    # pylint: disable=protected-access
    def get_origin_model(self, model_cls):
        """Return origin model even if proxy has been received."""
        return model_cls._meta.proxy_for_model or model_cls._meta.model

    def get_required_permissions(self, method, model_cls):
        """Add ability to define origin model even via proxy."""
        model_cls = self.get_origin_model(model_cls)
        return super().get_required_permissions(method, model_cls)

    def get_perms_list(self, view, obj=None):
        """Return permission list for action from backend and view."""
        perms_map_name = "perms_map_action"
        view_perms_map = getattr(view, perms_map_name, {})
        view_perms_list = view_perms_map.get(view.action)
        if view_perms_list is not None:
            return view_perms_list
        self.generate_perms_map_action(view)
        backend_perms_map = getattr(self, perms_map_name, {})
        return backend_perms_map.get(view.action) or []

    # pylint: disable=protected-access
    def get_required_action_permissions(self, view, model_cls, obj=None):
        """Given a model and an action, return the list of permission codes."""
        model_cls = self.get_origin_model(model_cls)
        kwargs = {
            "app_label": model_cls._meta.app_label,
            "model_name": model_cls._meta.model_name,
        }
        return [perm % kwargs if isinstance(perm, str) else perm for perm in self.get_perms_list(view, obj)]

    def user_has_action_perm(self, user, view, perm, obj=None):
        """Check if user has single permission for particular view action."""
        assert callable(perm) or isinstance(perm, str), "Permission must be function or string"

        if callable(perm):
            return perm(user, view, obj)

        return user.has_perm(perm)

    def get_model_cls(self, view):
        model_cls = getattr(view, "model_class", None)
        if not model_cls:
            queryset = self._queryset(view)
            assert queryset, "queryset or model_class is required"
            model_cls = queryset.model

        return model_cls

    def has_action_permission(self, request, view, obj=None):
        """Check action specific permissions ignoring custom method."""
        model_cls = self.get_model_cls(view)
        perms = self.get_required_action_permissions(view, model_cls, obj)
        return all(self.user_has_action_perm(request.user, view, perm, obj) for perm in perms)

    def has_permission(self, request, view):
        """Apply action permission without object and with ignoring method."""
        if getattr(view, "_ignore_model_permissions", False):
            return True

        if not request.user or (not request.user.is_authenticated and self.authenticated_users_only):
            return False

        if view.action in getattr(view, "permission_exclude_views", []):
            return True

        if getattr(view, "include_model_default_method_permission", None):
            model_cls = self.get_model_cls(view)
            perms = self.get_required_permissions(request.method, model_cls)
            result = all(self.user_has_action_perm(request.user, view, perm) for perm in perms)
            return result and self.has_action_permission(request, view)

        return self.has_action_permission(request, view)

    def has_object_permission(self, request, view, obj):
        """Apply action permission with object and with ignoring method."""
        return self.has_action_permission(request, view, obj)



import uuid
from collections.abc import Callable

import structlog
from django.http import HttpRequest, HttpResponse

from payla_utils.settings import payla_utils_settings

logger = structlog.get_logger(__name__)


def get_request_header(request, header_key, meta_key):
    if hasattr(request, "headers"):
        return request.headers.get(header_key)

    return request.META.get(meta_key)


class RequestMiddleware:
    """``RequestMiddleware`` adds request metadata to ``structlog``'s logger context automatically.
    Make sure to add it after AuthenticationMiddleware in your MIDDLEWARE setting to have the request.user available.
    >>> MIDDLEWARE = [
    ...     # ...
    ...     'payla_utils.middlewares.RequestMiddleware',
    ... ]
    """

    def __init__(self, get_response: Callable) -> None:
        self.get_response = get_response
        self._raised_exception = False

    def __call__(self, request: HttpRequest) -> HttpResponse:
        from ipware import get_client_ip

        request_id_header = payla_utils_settings.REQUEST_ID_HEADER or "X-Request-ID"

        request_id = get_request_header(
            request, request_id_header.lower(), f"HTTP_{request_id_header.upper().replace('-', '_')}"
        ) or str(uuid.uuid4())

        structlog.contextvars.bind_contextvars(request_id=request_id)
        self.bind_user_id(request)

        ip, _ = get_client_ip(request)
        structlog.contextvars.bind_contextvars(ip=ip)

        # add request host, method, and path
        structlog.contextvars.bind_contextvars(
            request_host=request.get_host(),
            request_method=request.method,
            request_path=request.path,
        )

        response = self.get_response(request)

        self.bind_user_id(request)

        structlog.contextvars.clear_contextvars()
        return response

    @staticmethod
    def bind_user_id(request: HttpRequest) -> None:
        if hasattr(request, "user") and request.user is not None:
            user_id = None
            if hasattr(request.user, "pk"):
                user_id = request.user.pk
                if isinstance(user_id, uuid.UUID):
                    user_id = str(user_id)
            structlog.contextvars.bind_contextvars(user_id=user_id)



# ruff: noqa: S603,S607

import subprocess
from io import StringIO

from django.core.management import BaseCommand, call_command


class Command(BaseCommand):
    help = "Make and compile messages for german language"
    requires_system_checks: list = []

    def handle(self, *args, **options):
        self.stdout.write("Making messages", style_func=self.style.SUCCESS)
        out = StringIO()
        call_command(
            "makemessages",
            locale=["de"],
            no_location=True,
            no_obsolete=True,
            extension=["html", "py", "txt", "js", "tpl"],
            stdout=out,
        )

        self.stdout.write(out.getvalue())

        # remove POT-Creation-Date from all .po files using sed
        subprocess.call(["find", ".", "-name", "*.po", "-exec", "sed", "-i", "-e", "/POT-Creation-Date/d", "{}", ";"])

        self.stdout.write("Compiling messages", style_func=self.style.SUCCESS)
        call_command("compilemessages", locale=["de"])



from __future__ import annotations

import logging
import logging.config

import structlog
from structlog import contextvars
from structlog.typing import EventDict
from structlog_sentry import SentryProcessor

from payla_utils.logging.default_configuration import get_default_logging_conf
from payla_utils.logging.sentry_logger import ignore_logger


class LoggerSetupError(Exception):
    pass


class LoggingConfigurator:
    def __init__(
        self,
        service_name: str,
        log_level: str,
        own_apps: list | None = None,
        config: dict | None = None,
        setup_logging_dict: bool = False,
    ):
        self.service_name = service_name
        self.log_level = log_level.upper()
        self.config = config
        self.own_apps = own_apps or []
        self.setup_logging_dict = setup_logging_dict

    @staticmethod
    def add_logger_name(logger: logging.Logger, method_name: str, event_dict: EventDict) -> EventDict:
        """
        Add the logger name to the event dict.
        """
        record = event_dict.get("_record")
        if record is None:
            event_dict["name"] = logger.name
        else:
            event_dict["name"] = record.name
        return event_dict

    def add_open_telemetry_spans(self, _, __, event_dict):
        from opentelemetry import trace
        from opentelemetry.trace import format_span_id, format_trace_id

        span = trace.get_current_span()
        if not span.is_recording():
            return event_dict

        ctx = span.get_span_context()
        parent = getattr(span, "parent", None)

        event_dict.update(
            {
                "span_id": format_span_id(ctx.span_id),
                "trace_id": format_trace_id(ctx.trace_id),
                "parent_span_id": None if not parent else format_span_id(parent.span_id),
            }
        )

        return event_dict

    def get_processors(self, tracing_enabled: bool) -> list:
        processors = [
            contextvars.merge_contextvars,
            self.add_logger_name,
            structlog.stdlib.add_log_level,
            structlog.stdlib.filter_by_level,
            structlog.processors.TimeStamper(fmt="iso"),
            structlog.stdlib.PositionalArgumentsFormatter(),
            structlog.processors.StackInfoRenderer(),
            structlog.dev.set_exc_info,
            structlog.processors.format_exc_info,
            structlog.processors.UnicodeDecoder(),
            structlog.processors.ExceptionPrettyPrinter(),
            structlog.stdlib.ExtraAdder(),
            SentryProcessor(event_level=logging.ERROR),
            structlog.processors.EventRenamer(to="message"),
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ]
        if tracing_enabled:
            processors.insert(3, self.add_open_telemetry_spans)
        return processors

    def _fix_sentry_logging(self, extra_loggers_names: list[str] | None = None):
        """
        Ignore loggers using struct log to avoid duplicated events sent to sentry
        """
        for logger_name in [app.split('.', maxsplit=1)[0] for app in self.own_apps] + (extra_loggers_names or []):
            ignore_logger(logger_name)

    def configure_structlog(
        self,
        custom_processors=None,
        formatter='json_formatter',
        formatter_std_lib='json_formatter',
        tracing_enabled=False,
        extra_loggers_names: list[str] | None = None,
    ):
        if self.setup_logging_dict:
            logger_init_config = self.config or get_default_logging_conf(
                log_level=self.log_level,
                own_apps=self.own_apps,
                formatter=formatter,
                formatter_std_lib=formatter_std_lib,
            )
            for extra_logger_name in extra_loggers_names or []:
                logger_init_config['loggers'][extra_logger_name] = {
                    "handlers": ["console"],
                    "level": self.log_level,
                    "propagate": False,
                }

            logging.config.dictConfig(logger_init_config)

        self._fix_sentry_logging(extra_loggers_names=extra_loggers_names)

        processors = custom_processors or self.get_processors(tracing_enabled=tracing_enabled)

        structlog.configure(
            processors=processors,
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=True,
        )



import os
from typing import Any

from admin_interface.cache import del_cached_active_theme
from admin_interface.models import Theme
from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.core.management.base import BaseCommand

from payla_utils.settings import payla_utils_settings


class Command(BaseCommand):
    """
    This management command will init environment based on the current env (local.dev, stage, playground and prod)

    - load fixtures on the first run (when the DB is empty)
    - setup custom theme for admin_interface
    - create user when not in prod if `LOCAL_DJANGO_ADMIN_PASSWORD` is set

    APP_NAME and ENVIRONMENT settings are required.
    """

    def add_arguments(self, parser):
        parser.add_argument('--env', type=str, default='', help='force which env to use')

    def handle(self, *args: Any, **options: Any):
        env = options.get('env') or payla_utils_settings.ENVIRONMENT
        self.setup_theme(env)
        self.setup_initial_fixtures(env=env)
        self.run_extra_commands()

    def setup_initial_fixtures(self, env: str) -> None:
        if not get_user_model().objects.all().exists():
            for fixture in payla_utils_settings.INITIAL_FIXTURES:
                call_command('loaddata', fixture)
            if not env.lower().startswith('prod') and payla_utils_settings.LOCAL_DJANGO_ADMIN_PASSWORD:
                # password has to be handed over as an environment variable
                os.environ.setdefault('DJANGO_SUPERUSER_PASSWORD', payla_utils_settings.LOCAL_DJANGO_ADMIN_PASSWORD)
                call_command('createsuperuser', '--username', 'admin', '--email', 'tools@payla.de', '--noinput')
        else:
            self.stdout.write(
                self.style.WARNING("Not importing fixtures as there seems to be data present. Please check.")
            )

    def setup_theme(self, env: str) -> None:
        base_title = payla_utils_settings.APP_NAME
        theme_settings = {
            'css_module_rounded_corners': True,
            'env_name': env.upper(),
            'title': f'{base_title} ({env.upper()})',
        }
        env_themes = payla_utils_settings.ENV_THEMES

        if env.lower() not in env_themes:
            self.stdout.write(
                self.style.WARNING(f'No theme settings for env {env.lower()} found. Skipping theme setup.')
            )
            return

        theme: Theme = Theme.objects.get_active()
        theme_settings.update(env_themes[env.lower()])
        # update theme name if set to APP_NAME key
        if theme_settings['title'] == 'APP_NAME':
            theme_settings['title'] = base_title
        for field_name, field_value in theme_settings.items():
            setattr(theme, field_name, field_value)
        theme.save()
        del_cached_active_theme()

    def run_extra_commands(self):
        for command in payla_utils_settings.RUN_EXTRA_COMMANDS:
            custom_command = command.split(' ')
            # handle custom commands with arguments
            if len(custom_command) > 1:
                new_command, *args = custom_command
                call_command(new_command, *args)
            else:
                call_command(command)



from datetime import datetime, timezone

from pythonjsonlogger import jsonlogger

from payla_utils.settings import payla_utils_settings


# This is straight from python-json-logger's documentation and will
# add a few useful default fields to each log message.
class LogFormatter(jsonlogger.JsonFormatter):
    def add_fields(self, log_record, record, message_dict):
        super().add_fields(log_record, record, message_dict)
        # add project name
        log_record['microservice'] = payla_utils_settings.MICROSERVICE_NAME
        if not log_record.get('timestamp'):
            now = datetime.now(tz=timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%fZ')
            log_record['timestamp'] = now
        if log_record.get('level'):
            log_record['level'] = log_record['level'].upper()
        else:
            log_record['level'] = record.levelname



from functools import wraps

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.http.request import split_domain_port
from django.http.response import Http404

from payla_utils.settings import payla_utils_settings


def only_internal_access(view_func):
    """
    Decorator for views that checks that the endpoint is only accessible via the direct IP Address.
    """

    @wraps(view_func)
    def _wrapped_view(request, *args, **kwargs):
        domain, __ = split_domain_port(request.get_host())
        server_ip = payla_utils_settings.SERVER_IP
        if not server_ip:
            raise Exception('SERVER_IP is required in order to use only_internal_access decorator')
        if domain == server_ip:
            return view_func(request, *args, **kwargs)
        raise Http404('Not found')

    return _wrapped_view


def staff_member_required_or_local_dev(
    view_func=None, redirect_field_name=REDIRECT_FIELD_NAME, login_url="admin:login"
):
    """
    Same as staff member required but allows local development
    """

    @wraps(view_func)
    def _wrapped_view(request, *args, **kwargs):
        if settings.ENVIRONMENT == "local.dev":
            return view_func(request, *args, **kwargs)
        return staff_member_required(redirect_field_name=redirect_field_name, login_url=login_url)(view_func)(
            request, *args, **kwargs
        )

    return _wrapped_view



from payla_utils.logging.default_configuration import get_default_logging_conf
from payla_utils.logging.logger import LoggingConfigurator

__all__ = ["LoggingConfigurator", "get_default_logging_conf"]



import json
from pathlib import Path
from typing import Any

import structlog
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AbstractUser, Group, Permission
from django.core.management.base import BaseCommand
from django.db import transaction
from pydantic import BaseModel

from payla_utils.settings import payla_utils_settings

logger = structlog.get_logger(__name__)

User = get_user_model()


# region: Schema classes for parsing users configurations from JSON file
class UserBasicInfoConfig(BaseModel):
    username: str
    first_name: str
    last_name: str
    email: str


class UserEnvConfig(BaseModel):
    is_superuser: bool
    is_staff: bool
    is_active: bool


class UserConfig(BaseModel):
    basic_info: UserBasicInfoConfig
    envs: dict[str, UserEnvConfig]
    groups: list[str]

    def to_minimal_dict(self, env_str: str) -> dict:
        basic_info = self.basic_info
        env_config = self.envs[env_str]
        return {
            "first_name": basic_info.first_name,
            "last_name": basic_info.last_name,
            "email": basic_info.email,
            "is_active": env_config.is_active,
            "is_staff": env_config.is_staff,
            "is_superuser": env_config.is_superuser,
        }


#  endregion


class Command(BaseCommand):
    """
    Django management command to set up groups, permissions, and users.

    This command performs the following functions:
        1. Adding Permissions: it reads a JSON file that defines groups and their respective permissions.
           For each group, the command creates it in the database (if it doesn't already exist) and assigns
           the specified permissions to it.
        2. Removing Unspecified Permissions: the command also checks for and removes any permissions
           currently assigned to the group but not specified in the JSON configuration. This ensures that
           each group only has the permissions explicitly defined in the configuration.
        3. Adding Users (if they don't already exist): it reads ANOTHER JSON file that contains essential data for
           creating users and assigning them to appropriate groups. User hashed passwords are read from an environment
           variable that's taken from AWS secret manager.
        4. Removing users from groups they should not belong to, correcting potential mistakes made via Admin.
        5. Clearing User-Level Permissions: users should receive permissions only via their assigned groups.

    Example 1: JSON structure for group permissions(core is the app_label, bank is the model name, view_bank is the
    permission codename):

    {
        "core": {
            "bank": {
                "view_bank": [
                    {"group": "customer_service_level_1", "envs": ["playground", "prod", "stage"]},
                ],
                "change_bank": [
                    {"group": "finance", "envs": ["prod", "stage"]},
                ]
            }
        }
    }

    Example 2: JSON structure for group users:
    [
        {
            "basic_info": {
                "username": "john.doe@payla.de",
                "first_name": "John",
                "last_name": "Doe",
                "email": "john.doe@payla.de"
            },
            "envs": {
                "stage": {
                    "is_superuser": false,
                    "is_staff": true,
                    "is_active": true,
                },
                "playground": {
                    "is_superuser": false,
                    "is_staff": true,
                    "is_active": true,
                },
                "prod": {
                    "is_superuser": false,
                    "is_staff": true,
                    "is_active": true,
                }
            },
            "groups": [
                "Backoffice",
                "engineering",
                "core"
            ]
        }
    ]

    Example 3: AWS secret manager environment variable for user passwords (per environment):

    USER_PASSWORDS = {
        "username1": "pbkdf2_sha256$870000$YPcJGsFOSJM3uEoNbSePfX$tWi4dy66J9I7BOV6De04AbaVD8g9UBGH9wzatGYSt2c=",
        "username2": "pbkdf2_sha256$870000$YPcJGsFOSJM3uEoNbSePfX$tWi4dy66J9I7BOV6De04AbaVD8g9UBGH9wzatGYSt2c=",
    }

    How to create the hash for the password:

    ```python
    from django.contrib.auth.hashers import make_password
    make_password('your_password_here')
    ```

    """

    help = 'Syncs groups permissions and users based on JSON configuration files.'

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.current_env = ''
        self.groups_permissions: dict[str, list[str]] = {}

    def assign_group_permissions(self, groups_permissions_config: dict):
        for app_label, app_data in groups_permissions_config.items():
            for model_name, model_data in app_data.items():
                for permission_codename, permission_data in model_data.items():
                    self.process_group_permission(app_label, model_name, permission_codename, permission_data)

    def process_group_permission(
        self, app_label: str, model_name: str, permission_codename: str, permission_data: dict
    ):
        permission_code = f"{permission_codename}.{app_label}.{model_name}"

        try:
            permission = Permission.objects.get_by_natural_key(permission_codename, app_label, model_name)
        except Permission.DoesNotExist:
            logger.exception("Permission %s not found!", permission_code)
            return

        for permission_group in permission_data:
            envs = permission_group.get('envs', [])
            groups_names = (
                permission_group['groups']
                if isinstance(permission_group['groups'], list)
                else [permission_group['group']]
            )

            if self.current_env not in envs:
                logger.info(
                    "Skipping permission %s for group(s) %s in environment %s",
                    permission_codename,
                    groups_names,
                    self.current_env,
                )
                continue

            for group_name in groups_names:
                # Create group
                group = Group.objects.get_or_create(name=group_name)[0]

                self.groups_permissions.setdefault(group_name, [])

                # Add permission to group
                group.permissions.add(permission)
                logger.info("Added permission %s to group %s", permission_codename, group_name)
                self.groups_permissions[group_name].append(permission_codename)

    def remove_unwanted_group_permissions(self) -> None:
        # Remove unused permissions from group
        for group_name, permissions in self.groups_permissions.items():
            group = Group.objects.get(name=group_name)
            removed_permissions = 0
            for permission in group.permissions.all():
                if permission.codename not in permissions:
                    logger.info("Removing permission %s from group %s", permission.codename, group_name)
                    group.permissions.remove(permission)
                    removed_permissions += 1
            logger.info("Removed %s permissions from group %s", removed_permissions, group_name)

    def sync_group_permissions(self) -> None:
        if not payla_utils_settings.GROUPS_PERMISSIONS_FILE_PATH:
            logger.warning("Skipping groups permissions sync. Missing GROUPS_PERMISSIONS_FILE_PATH in settings.")
            return

        with Path.open(payla_utils_settings.GROUPS_PERMISSIONS_FILE_PATH, encoding='utf-8') as f:
            groups_permissions_config = json.loads(f.read())

        self.assign_group_permissions(groups_permissions_config)
        self.remove_unwanted_group_permissions()

    def update_or_create_user(self, user_config: UserConfig, password: str) -> AbstractUser:
        basic_info_config = user_config.basic_info

        user_db = User.objects.filter(username=basic_info_config.username).first()

        if user_db:
            for attr, value in user_config.to_minimal_dict(self.current_env).items():
                setattr(user_db, attr, value)
        else:
            user_db = User.objects.create_user(
                username=basic_info_config.username,
                **user_config.to_minimal_dict(self.current_env),
            )

        if password:
            user_db.password = password
        else:
            user_db.set_unusable_password()

        user_db.save()

        return user_db

    def add_user_to_groups(self, user: AbstractUser, groups: list[str]) -> None:
        for group_name in groups:
            group, __ = Group.objects.get_or_create(name=group_name)
            user.groups.add(group)

    def remove_unwanted_groups(self, user: AbstractUser, desired_groups: list[str]) -> None:
        for group in user.groups.all():
            if group.name not in desired_groups:
                user.groups.remove(group)

    def clear_user_level_permissions(self, user: AbstractUser) -> None:
        user_level_permissions = user.get_user_permissions()
        if user_level_permissions:
            user.user_permissions.clear()
            logger.info(
                'User-level permissions have been cleared for: %s',
                user.username,
                user_level_permissions=user_level_permissions,
            )

    def sync_group_users(self) -> None:
        config_file = payla_utils_settings.GROUPS_USERS_FILE_PATH
        if not config_file:
            logger.warning("Skipping groups users sync. Missing GROUPS_USERS_FILE_PATH in settings.")
            return

        passwords_map = payla_utils_settings.USER_PASSWORDS
        if not passwords_map:
            logger.warning("Skipping groups users sync. Missing USER_PASSWORDS in settings.")
            return

        with Path.open(config_file, encoding='utf-8') as f:
            users_configs_data = json.loads(f.read())

        for user_config_data in users_configs_data:
            logger.info("Groups users sync: Processing user %s", user_config_data['basic_info']['username'])
            user_config = UserConfig(**user_config_data)

            if self.current_env not in user_config.envs:
                logger.info(
                    "Groups users sync: Skipping user %s in environment %s",
                    user_config.base.username,
                    self.current_env,
                )
                continue

            password = passwords_map.get(user_config.basic_info.username)

            user_db = self.update_or_create_user(user_config, password)
            self.add_user_to_groups(user_db, user_config.groups)
            self.remove_unwanted_groups(user_db, user_config.groups)
            self.clear_user_level_permissions(user_db)

    @transaction.atomic
    def handle(self, *args: Any, **options: Any) -> str | None:
        self.current_env = 'stage' if settings.ENVIRONMENT == 'local.dev' else settings.ENVIRONMENT

        self.sync_group_permissions()

        self.sync_group_users()

        return None



from django.contrib.auth import get_user_model


def run():
    test_user = get_user_model().objects.filter(username="test_user").first()
    if not test_user:
        test_user = get_user_model()(username="test_user")
        test_user.set_unusable_password()
        test_user.save()
